@@ -3,18 +3,22 @@
 /**
  * SQL import plugin for phpMyAdmin
  *
- * @version $Id: sql.php 11051 2008-01-14 12:59:42Z lem9 $
+ * @version $Id$
+ * @package phpMyAdmin-Import
  */
+if (! defined('PHPMYADMIN')) {
+    exit;
+}
 
 /**
  *
  */
 if (isset($plugin_list)) {
     $plugin_list['sql'] = array(
-        'text' => 'strSQL',
-        'extension' => 'sql',
-        'options_text' => 'strOptions',
-        );
+        'text'          => 'strSQL',
+        'extension'     => 'sql',
+        'options_text'  => 'strOptions',
+    );
     $compats = PMA_DBI_getCompatibilities();
     if (count($compats) > 0) {
         $values = array();
@@ -22,8 +26,28 @@ if (isset($plugin_list)) {
             $values[$val] = $val;
         }
         $plugin_list['sql']['options'] = array(
-            array('type' => 'select', 'name' => 'compatibility', 'text' => 'strSQLCompatibility', 'values' => $values, 'doc' => array('manual_MySQL_Database_Administration', 'Server_SQL_mode'))
-            );
+            array(
+                'type'      => 'select',
+                'name'      => 'compatibility',
+                'text'      => 'strSQLCompatibility',
+                'values'    => $values,
+                'doc'       => array(
+                    'manual_MySQL_Database_Administration',
+                    'Server_SQL_mode',
+                ),
+            ),
+            array(
+                'type' => 'bool', 
+                'name' => 'no_auto_value_on_zero', 
+                'text' => 'strDoNotAutoIncrementZeroValues',
+                'doc'       => array(
+                    'manual_MySQL_Database_Administration',
+                    'Server_SQL_mode',
+                    'sqlmode_no_auto_value_on_zero'
+                ),
+
+            ),
+        );
     }
 
     /* We do not define function when plugin is just queried for information above */
@@ -36,17 +60,37 @@ $sql = '';
 $start_pos = 0;
 $i = 0;
 $len= 0;
+$big_value = 2147483647;
+$delimiter_keyword = 'DELIMITER '; // include the space because it's mandatory
+$length_of_delimiter_keyword = strlen($delimiter_keyword);
+
 if (isset($_POST['sql_delimiter'])) {
     $sql_delimiter = $_POST['sql_delimiter'];
 } else {
     $sql_delimiter = ';';
 }
 
-// Handle compatibility option
-if (isset($_REQUEST['sql_compatibility'])) {
-    PMA_DBI_try_query('SET SQL_MODE="' . $_REQUEST['sql_compatibility'] . '"');
+// Handle compatibility options
+$sql_modes = array();
+if (isset($_REQUEST['sql_compatibility']) && 'NONE' != $_REQUEST['sql_compatibility']) {
+    $sql_modes[] = $_REQUEST['sql_compatibility'];
 }
-while (!($finished && $i >= $len) && !$error && !$timeout_passed) {
+if (isset($_REQUEST['sql_no_auto_value_on_zero'])) {
+    $sql_modes[] = 'NO_AUTO_VALUE_ON_ZERO';
+}
+if (count($sql_modes) > 0) {
+    PMA_DBI_try_query('SET SQL_MODE="' . implode(',', $sql_modes) . '"');
+}
+unset($sql_modes);
+
+/**
+ * will be set in PMA_importGetNextChunk()
+ *
+ * @global boolean $GLOBALS['finished']
+ */
+$GLOBALS['finished'] = false;
+
+while (!($GLOBALS['finished'] && $i >= $len) && !$error && !$timeout_passed) {
     $data = PMA_importGetNextChunk();
     if ($data === FALSE) {
         // subtract data we didn't handle yet and stop processing
@@ -60,73 +104,47 @@ while (!($finished && $i >= $len) && !$error && !$timeout_passed) {
         // free memory
         unset($data);
         // Do not parse string when we're not at the end and don't have ; inside
-        if ((strpos($buffer, $sql_delimiter, $i) === FALSE) && !$finished)  {
+        if ((strpos($buffer, $sql_delimiter, $i) === FALSE) && !$GLOBALS['finished'])  {
             continue;
         }
     }
     // Current length of our buffer
     $len = strlen($buffer);
-    // prepare an uppercase copy of buffer for PHP < 5
-    // outside of the loop
-    if (PMA_PHP_INT_VERSION < 50000) {
-        $buffer_upper = strtoupper($buffer);
-    }
+
     // Grab some SQL queries out of it
-     while ($i < $len) {
+    while ($i < $len) {
         $found_delimiter = false;
-        // Find first interesting character, several strpos seem to be faster than simple loop in php:
-        //while (($i < $len) && (strpos('\'";#-/', $buffer[$i]) === FALSE)) $i++;
-        //if ($i == $len) break;
-        $oi = $i;
-        $big_value = 2147483647;
-        $first_quote = strpos($buffer, '\'', $i);
-        if ($first_quote === FALSE) {
-            $first_quote = $big_value;
-        }
-        $p2 = strpos($buffer, '"', $i);
-        if ($p2 === FALSE) {
-            $p2 = $big_value;
+        // Find first interesting character
+        $old_i = $i;
+        // this is about 7 times faster that looking for each sequence i
+        // one by one with strpos()
+        if (preg_match('/(\'|"|#|-- |\/\*|`|(?i)(?<![A-Z0-9_])' . $delimiter_keyword . ')/', $buffer, $matches, PREG_OFFSET_CAPTURE, $i)) {
+            // in $matches, index 0 contains the match for the complete
+            // expression but we don't use it
+            $first_position = $matches[1][1];
+        } else {
+            $first_position = $big_value;
         }
         /**
-         * @todo it's a shortcoming to look for a delimiter that might be
+         * @todo we should not look for a delimiter that might be
          *       inside quotes (or even double-quotes)
          */
+        // the cost of doing this one with preg_match() would be too high
         $first_sql_delimiter = strpos($buffer, $sql_delimiter, $i);
         if ($first_sql_delimiter === FALSE) {
             $first_sql_delimiter = $big_value;
         } else {
             $found_delimiter = true;
         }
-        $p4 = strpos($buffer, '#', $i);
-        if ($p4 === FALSE) {
-            $p4 = $big_value;
-        }
-        $p5 = strpos($buffer, '--', $i);
-        if ($p5 === FALSE || $p5 >= ($len - 2) || $buffer[$p5 + 2] > ' ') {
-            $p5 = $big_value;
-        }
-        $p6 = strpos($buffer, '/*', $i);
-        if ($p6 === FALSE) {
-            $p6 = $big_value;
-        }
-        $p7 = strpos($buffer, '`', $i);
-        if ($p7 === FALSE) {
-            $p7 = $big_value;
-        }
-        // catch also "delimiter"
-        if (PMA_PHP_INT_VERSION >= 50000) {
-            $p8 = stripos($buffer, 'DELIMITER', $i);
-        } else {
-            $p8 = strpos($buffer_upper, 'DELIMITER', $i);
-        }
-        if ($p8 === FALSE || $p8 >= ($len - 11) || $buffer[$p8 + 9] > ' ') {
-            $p8 = $big_value;
-        }
-        $i = min ($first_quote, $p2, $first_sql_delimiter, $p4, $p5, $p6, $p7, $p8);
-        unset($first_quote, $p2, $p4, $p5, $p6, $p7, $p8);
+
+        // set $i to the position of the first quote, comment.start or delimiter found
+        $i = min($first_position, $first_sql_delimiter);
+
         if ($i == $big_value) {
-            $i = $oi;
-            if (!$finished) {
+            // none of the above was found in the string
+
+            $i = $old_i;
+            if (!$GLOBALS['finished']) {
                 break;
             }
             // at the end there might be some whitespace...
@@ -152,7 +170,7 @@ while (!($finished && $i >= $len) && !$error && !$timeout_passed) {
                 // No quote? Too short string
                 if ($pos === FALSE) {
                     // We hit end of string => unclosed quote, but we handle it as end of query
-                    if ($finished) {
+                    if ($GLOBALS['finished']) {
                         $endq = TRUE;
                         $i = $len - 1;
                     }
@@ -176,7 +194,7 @@ while (!($finished && $i >= $len) && !$error && !$timeout_passed) {
             }
             $i++;
             // Aren't we at the end?
-            if ($finished && $i == $len) {
+            if ($GLOBALS['finished'] && $i == $len) {
                 $i--;
             } else {
                 continue;
@@ -186,25 +204,29 @@ while (!($finished && $i >= $len) && !$error && !$timeout_passed) {
         // Not enough data to decide
         if ((($i == ($len - 1) && ($ch == '-' || $ch == '/'))
           || ($i == ($len - 2) && (($ch == '-' && $buffer[$i + 1] == '-')
-            || ($ch == '/' && $buffer[$i + 1] == '*')))) && !$finished) {
+            || ($ch == '/' && $buffer[$i + 1] == '*')))) && !$GLOBALS['finished']) {
             break;
         }
 
         // Comments
         if ($ch == '#'
-                || ($i < ($len - 1) && $ch == '-' && $buffer[$i + 1] == '-' && (($i < ($len - 2) && $buffer[$i + 2] <= ' ') || ($i == ($len - 1) && $finished)))
-                || ($i < ($len - 1) && $ch == '/' && $buffer[$i + 1] == '*')
+         || ($i < ($len - 1) && $ch == '-' && $buffer[$i + 1] == '-'
+          && (($i < ($len - 2) && $buffer[$i + 2] <= ' ')
+           || ($i == ($len - 1)  && $GLOBALS['finished'])))
+         || ($i < ($len - 1) && $ch == '/' && $buffer[$i + 1] == '*')
                 ) {
             // Copy current string to SQL
             if ($start_pos != $i) {
                 $sql .= substr($buffer, $start_pos, $i - $start_pos);
             }
             // Skip the rest
-            $j = $i;
+            $start_of_comment = $i;
+            // do not use PHP_EOL here instead of "\n", because the export 
+            // file might have been produced on a different system
             $i = strpos($buffer, $ch == '/' ? '*/' : "\n", $i);
             // didn't we hit end of string?
             if ($i === FALSE) {
-                if ($finished) {
+                if ($GLOBALS['finished']) {
                     $i = $len - 1;
                 } else {
                     break;
@@ -212,21 +234,13 @@ while (!($finished && $i >= $len) && !$error && !$timeout_passed) {
             }
             // Skip *
             if ($ch == '/') {
-                // Check for MySQL conditional comments and include them as-is
-                if ($buffer[$j + 2] == '!') {
-                    $comment = substr($buffer, $j + 3, $i - $j - 3);
-                    if (preg_match('/^[0-9]{5}/', $comment, $version)) {
-                        if ($version[0] <= PMA_MYSQL_INT_VERSION) {
-                            $sql .= substr($comment, 5);
-                        }
-                    } else {
-                        $sql .= $comment;
-                    }
-                }
                 $i++;
             }
             // Skip last char
             $i++;
+            // We need to send the comment part in case we are defining
+            // a procedure or function and comments in it are valuable
+            $sql .= substr($buffer, $start_of_comment, $i - $start_of_comment);
             // Next query part will start here
             $start_pos = $i;
             // Aren't we at the end?
@@ -236,29 +250,37 @@ while (!($finished && $i >= $len) && !$error && !$timeout_passed) {
                 continue;
             }
         }
-       // Change delimiter, if redefined, and skip it (don't send to server!)
-       if ((strtoupper(substr($buffer, $i, 9)) == "DELIMITER") && ($buffer[$i + 9] <= ' ') && ($i<$len-11) && (!(strpos($buffer,"\n",$i+11)===FALSE))) {
-           $new_line_pos = strpos($buffer, "\n", $i + 10);
-           $sql_delimiter = substr($buffer, $i+10, $new_line_pos - $i -10);
-           $i= $new_line_pos + 1;
+        // Change delimiter, if redefined, and skip it (don't send to server!)
+        if (strtoupper(substr($buffer, $i, $length_of_delimiter_keyword)) == $delimiter_keyword
+         && ($i + $length_of_delimiter_keyword < $len)) {
+             // look for EOL on the character immediately after 'DELIMITER '
+             // (see previous comment about PHP_EOL)
+           $new_line_pos = strpos($buffer, "\n", $i + $length_of_delimiter_keyword);
+           // it might happen that there is no EOL
+           if (FALSE === $new_line_pos) {
+               $new_line_pos = $len;
+           }
+           $sql_delimiter = substr($buffer, $i + $length_of_delimiter_keyword, $new_line_pos - $i - $length_of_delimiter_keyword);
+           $i = $new_line_pos + 1;
            // Next query part will start here
            $start_pos = $i;
            continue;
         }
 
         // End of SQL
-        if ($found_delimiter || ($finished && ($i == $len - 1))) {
+        if ($found_delimiter || ($GLOBALS['finished'] && ($i == $len - 1))) {
             $tmp_sql = $sql;
             if ($start_pos < $len) {
                 $length_to_grab = $i - $start_pos;
-                if (!$found_delimiter) {
+
+                if (! $found_delimiter) {
                     $length_to_grab++;
                 }
                 $tmp_sql .= substr($buffer, $start_pos, $length_to_grab);
                 unset($length_to_grab);
             }
             // Do not try to execute empty SQL
-            if (!preg_match('/^([\s]*;)*$/', trim($tmp_sql))) {
+            if (! preg_match('/^([\s]*;)*$/', trim($tmp_sql))) {
                 $sql = $tmp_sql;
                 PMA_importRunQuery($sql, substr($buffer, 0, $i + strlen($sql_delimiter)));
                 $buffer = substr($buffer, $i + strlen($sql_delimiter));
@@ -268,8 +290,8 @@ while (!($finished && $i >= $len) && !$error && !$timeout_passed) {
                 $i = 0;
                 $start_pos = 0;
                 // Any chance we will get a complete query?
-                //if ((strpos($buffer, ';') === FALSE) && !$finished) {
-                if ((strpos($buffer, $sql_delimiter) === FALSE) && !$finished) {
+                //if ((strpos($buffer, ';') === FALSE) && !$GLOBALS['finished']) {
+                if ((strpos($buffer, $sql_delimiter) === FALSE) && !$GLOBALS['finished']) {
                     break;
                 }
             } else {
@@ -277,7 +299,6 @@ while (!($finished && $i >= $len) && !$error && !$timeout_passed) {
                 $start_pos = $i;
             }
         }
-
     } // End of parser loop
 } // End of import loop
 // Commit any possible data in buffers
